#ifndef UTIL_H
#define UTIL_H

#include <vector>
#include <iostream>
#include <utility>
#include <string>
#include <random>
#include <chrono>

// Some utility functions we may need. These are used in the lecture
// but we did not go over them.

// Values and functions for bookkeeping
extern int n_comparisons_;
extern int n_swaps_;
void ReportStats();
void ResetStats();

// A "fancier" integer that keeps track of number of comparisons and swaps.
class FancyInt {
 public:
  int value_;

  FancyInt(int value) : value_(value) {}
  FancyInt() : value_(0) {}

  // overloaded comparison operators to keep count of number of comparisons
  bool operator<(const FancyInt& that) const {
    n_comparisons_++;
    return this->value_ < that.value_;
  }
  bool operator>(const FancyInt& that) const {
    return that < *this;
  }
  bool operator<=(const FancyInt& that) const {
    return !(*this > that);
  }
  bool operator>=(const FancyInt& that) const {
    return !(*this < that);
  }
  // Overload operator<< to print fancy ints
  friend std::ostream& operator<<(std::ostream& out, const FancyInt& n) {
    return out << n.value_;
  }
  // Overload std::swap using argument-dependent lookup (ADL).
  //
  // We use `friend` to define std::swap inside the body of this class. If
  // you are more curious about this, feel free to ask about it in GauchoSpace.
  // It is not a part of what we are covering in this course.
  friend void swap(FancyInt& a, FancyInt& b) {
    n_swaps_++;
    std::swap(a.value_, b.value_);
  }
};

// A large object with costly swap and comparison. We implement this by holding
// a string inside. The constructor using an `int` is for building random
// vectors easily with `RandomVector`.
//
// Besides holding a string, its implementation is pretty much the same.
class LargeObject {
public:
  std::string value_;

  LargeObject(int value) : value_(std::to_string(value)) {}
  LargeObject() : value_("") {}

  // overloaded comparison operators to keep count of number of comparisons
  bool operator<(const LargeObject& that) const {
    n_comparisons_++;
    return this->value_ < that.value_;
  }
  bool operator>(const LargeObject& that) const {
    return that < *this;
  }
  bool operator<=(const LargeObject& that) const {
    return !(*this > that);
  }
  bool operator>=(const LargeObject& that) const {
    return !(*this < that);
  }
  // Overload operator<<
  friend std::ostream& operator<<(std::ostream& out, const LargeObject& n) {
    return out << n.value_;
  }
  // Overload std::swap using argument-dependent lookup (ADL)
  friend void swap(LargeObject& a, LargeObject& b) {
    n_swaps_++;
    std::swap(a.value_, b.value_);
  }
};

// A _very crude_ timer implementation. Do not use this in production code. It uses
// std::chrono for timekeeping, you can look up that library's documentation if you
// want but the implementation of this class is outside the scope of CS 32.
class Timer {
  using duration = std::chrono::high_resolution_clock::duration;
  
  std::chrono::high_resolution_clock::time_point tick_;
   duration elapsed_time_;
public:
  Timer() : tick_(std::chrono::high_resolution_clock::now()), elapsed_time_(0) {}
  
  void Tick() {
    tick_ = std::chrono::high_resolution_clock::now();
  }
  
  void Tock() {
    auto tock = std::chrono::high_resolution_clock::now();
    elapsed_time_ += tock - tick_;
  }

  // Resets the tick and the elapsed time
  void Reset() {
    tick_ = std::chrono::high_resolution_clock::now();
    elapsed_time_ = duration(0);
  }
  
  
  // Returns the total elapsed time between ticks and tocks as
  // milliseconds.
  double ElapsedTime() const {
    return double(elapsed_time_.count()) / (double(duration::period::den) / 1000);
  }
};

template <class T>
void PrintVector(const std::vector<T>& v) {
  bool first = true;
  for (const T &x : v) {
    if (first) {
      first = false;
    } else {
      std::cout << ", ";
    }
    std::cout << x;
  }
  std::cout << "\n";
}

// A function that returns a random 32-bit integer.
int32_t RandomInt();

template<typename T>
std::vector<T> RandomVector(size_t size) {
  std::vector<T> result;
  result.reserve(size);
  for (; size > 0; --size) {
    result.push_back(T(RandomInt()));
  }
  return result;
}


#endif // UTIL_H
