#include <algorithm>
#include <iostream>
#include <vector>
#include <utility>
#include "Util.h"

using namespace std;

// Standard library provides a hashing function for most types: std::hash
// We will use it by default

template<class Key, class Value, class Hasher = std::hash<Key>>
class HashMap {
public:
  using Cell = std::pair<Key, Value>;
private:
  // Some constants we will need

  // How much the table will grow when re-hashing
  static constexpr double GROWTH_FACTOR = 1.5;
  
  // The backing array, store pointers so we can check if there is a value in a
  // cell. There are better and more efficient ways to do this in modern C++, we
  // will cover some of them later in the quarter.
  vector<Cell *> table;

  // How many elements are in the table
  size_t n_elements;

  // The hash function instance we are going to use
  Hasher hash_fn;

  int rehash_depth = 0;

  // The core function of the hash table. If `key` is in the table, returns the
  // index of the corresponding cell, if `key` is not in the table; then returns
  // where `key` should be inserted. If the table is full, then returns -1.
  ssize_t LocateCell(const Key & key) const {
    size_t hash = hash_fn(key);
    size_t index = hash % table.size();

    Cell * cell = table[index];
    if (cell && cell->first != key) {
      return Probe(key, hash, index);
    }
    return index;
  }

  // Probing mechanism
  ssize_t Probe(const Key & key, size_t hash, const size_t original_index) const {
    // Linear probing
    // for (size_t index = original_index; index < table.size(); ++index) {
    //   if (table[index] == nullptr || table[index]->first == key) {
    //     return index;
    //   }
    // }
    // // Reached the end of the table, scan from the beginning
    // for (size_t index = 0; index < original_index; ++index) {
    //   if (table[index] == nullptr || table[index]->first == key) {
    //     return index;
    //   }
    // }

    // quadratic probing, this is a slightly inefficient implementation
    for (size_t i = 1; i < table.size(); ++i) {
      size_t index = (original_index + i * i) % table.size();
      if (table[index] == nullptr || table[index]->first == key) {
        return index;
      }
    }

    return -1;
  }

  // Grow the table by `GROWTH_FACTOR` and rehash the keys.
  void Rehash() {
    ++rehash_depth;
    // Effectively, move the old table out and allocate a new, larger table
    vector<Cell *> tmp_table(table.size() * GROWTH_FACTOR);
    swap(table, tmp_table);
    cout << string(rehash_depth, ' ') << "[ resizing, new capacity: " << table.size() << endl;
    n_elements = 0;
    size_t correct_n_elements = 0;
    // Move the cells to the new table, this triggers computing a new hash
    // because the size of the table has changed.
    for (auto cell : tmp_table) {
      if (cell) {
        insert(cell);
        correct_n_elements++;
      }
    }
    n_elements = correct_n_elements;
    std::cout << string(rehash_depth, ' ') << "] size: " << size() << " done\n";
    --rehash_depth;
  }

public:
  // The big 3
  explicit HashMap(size_t starting_capacity = 7) :
    table(starting_capacity),
    n_elements(0),
    hash_fn(Hasher())
  {}

  // delete these to make the class non-copyable
  HashMap(const HashMap& that) = delete;
  const HashMap& operator=(const HashMap&) = delete;

  // The destructor
  ~HashMap() {
    for (auto & cell : table) {
      delete cell;
    }
  }

  // Returns the cell corresponding to the key if there is one. Otherwise,
  // returns nullptr.
  Cell * find(const Key& key) const {
    auto index = LocateCell(key);
    if (index != -1) {
      return table[index];
    }
    return nullptr;
  }

  // This version of insert will copy the key and the value, so we take them as
  // const references.
  void insert(const Key& key, const Value& value) {
    auto index = LocateCell(key);
    while (index == -1) {
      Rehash();
      index = LocateCell(key);
    }

    if (table[index]) {
      // update existing value
      table[index]->second = value;
    } else {
      // create new cell
      table[index] = new Cell{key, value};
      n_elements++;
    }
  }

  // This version of insert will take ownership of the `Cell` and not do any
  // copying. It aborts the program if the key is already in the table.
  void insert(Cell * cell) {
    auto index = LocateCell(cell->first);
    while (index == -1) {
      Rehash();
      index = LocateCell(cell->first);
    }

    assert(table[index] == nullptr);

    table[index] = cell;
    n_elements++;
  }

  size_t size() const {
    return n_elements;
  }

  size_t capacity() const {
    return table.size();
  }

  // Iterator interface
  class Iterator {
  public:
    Iterator operator+(ssize_t n) const {
      ssize_t index = this->index;
      // go left until the beginning
      while (n < 0 && index >= 0) {
        // find the next non-empty cell
        do {
          index--;
        } while (index >= 0 && map->table[index] == nullptr);
        n++;
      }

      if (n == 0 || index < 0) {
        return Iterator(*map, index);
      }
        
      // go right until the end, this loop and the loop above are mutually
      // exclusive.
      while (n > 0 && index < map->table.size()) {
        // find the next non-empty cell
        do {
          index++;
        } while (index < map->table.size() && map->table[index] == nullptr);
        n--;
      }

      return Iterator(*map, index);
    }
    Iterator operator-(ssize_t n) const {
      return *this + (- n);
    }
    Iterator& operator++() {
      *this = *this + 1;
      return *this;
    }
    Cell& operator*() {
      assert(index > 0 && index < map->table.size() && "The iterator is out of bounds");
      return *map->table[index];
    }

    // comparison operators
    bool operator==(const Iterator& that) const {
      return map == that.map && index == that.index;
    }
    bool operator!=(const Iterator& that) const {
      return !(*this == that);
    }

    Iterator(HashMap &map, ssize_t index) : map(&map), index(index) {}
    Iterator() = delete;
  private:
    // the parent map
    HashMap* map;
    size_t index;
  };

  Iterator begin() {
    // check if the first cell is not null, if so return it; otherwise find the
    // next full cell
    if (table[0]) {
      return Iterator(*this, 0);
    }
    if (table.size() == 0) {
      // begin == end in this case
      return end();
    }
    // an implementation hack to find the first non-empty cell
    return Iterator(*this, 0) + 1;
  }

  Iterator end() {
    return Iterator(*this, table.size());
  }
};

int main(int argc, char * argv[]) {
  HashMap<int, int> h;

  std::vector<int> numbers{100,200,300,400,500};

  // insert n -> n^2 into the map
  for (int n : numbers) {
    h.insert(n, n*n);
    cout << "inserted " << n << " size: " << h.size() << " capacity: " << h.capacity() << endl;
  }

  // find and print the elements
  for (auto& cell : h) {
    cout << cell.first << " -> " << cell.second << endl;
  }

  for (int i = 0; i < 100; ++i) {
    unsigned int n = RandomInt();
    h.insert(n, n*n);
  }

  return 0;
}
